import json
from collections import OrderedDict

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

ALGORITHM = 'DCI-HMAC-SHA256'


class AuthRequest(object):
    def __init__(self, method='GET', endpoint='/', payload=None, headers=None, params=None):
        """
        AuthRequest object contains shortcut methods used by dciauth Signature object.

        :param method (str): http method. Uppercase value in GET/POST/PUT/DELETE (default: GET)
        :param endpoint (str): http endpoint (default: '/')
        :param payload (dict): http payload (default: {})
        :param headers (dict): http headers (default: {})
        :param params (dict): url parameters (default: {})
        """

        self.algorithm = ALGORITHM
        self.method = method.upper()
        self.endpoint = endpoint
        self.payload = payload or {}
        self.headers = self._lower(headers or {})
        self.filtered_headers = self._filter_headers(self.headers)
        self.params = params or {}

    def build_headers(self, client_type, client_id, signature):
        auth_string_format = ("{dci_algorithm} Credential={client_type}/{client_id}, "
                              "SignedHeaders={signed_headers}, Signature={signature},")
        headers = self.headers.copy()
        headers['authorization'] = auth_string_format.format(
            dci_algorithm=self.algorithm,
            client_type=client_type,
            client_id=client_id,
            signed_headers=self.get_signed_headers_string(),
            signature=signature)
        return headers

    def add_header(self, key, value):
        self.headers[key] = value

    def get_headers_string(self):
        headers_string = ''
        for key, value in self._order_dict(self.filtered_headers).items():
            headers_string += '%s:%s\n' % (key, value)
        return headers_string

    def get_signed_headers_string(self):
        headers = self._order_dict(self.filtered_headers).keys()
        return ';'.join(headers)

    def get_query_string(self):
        return urlencode(self._order_dict(self.params))

    def get_payload_string(self):
        if self.payload:
            return json.dumps(self._order_dict(self.payload))
        return ''

    def get_client_info(self):
        authorization_header = self.headers.get('authorization')
        info = {}
        if authorization_header:
            credentials = self.find_between(authorization_header, 'Credential=', ',').split('/')
            info['client_type'] = credentials[0]
            info['client_id'] = credentials[1]
            info['signature'] = self.find_between(authorization_header, 'Signature=', ',')
        return info

    @staticmethod
    def find_between(string, first, last):
        try:
            start = string.index(first) + len(first)
            end = string.index(last, start)
            return string[start:end]
        except ValueError:
            return ""

    @staticmethod
    def _lower(headers):
        return {key.lower(): value for key, value in headers.items()}

    @staticmethod
    def _order_dict(dictionary):
        return OrderedDict(sorted(dictionary.items(), key=lambda k: k[0]))

    def _filter_headers(self, headers):
        authorization_header = headers.get('authorization')
        if not authorization_header:
            return headers
        signed_headers = self.find_between(authorization_header, 'SignedHeaders=', ',').split(';')
        return {key: value for key, value in headers.items() if key in signed_headers}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import setuptools

from dciauth import version

root_dir = os.path.dirname(os.path.abspath(__file__))
readme = open(os.path.join(root_dir, 'README.md')).read()

setuptools.setup(
    name='dciauth',
    version=version.__version__,
    packages=['dciauth'],
    author='Distributed CI team',
    author_email='distributed-ci@redhat.com',
    description='DCI authentication module used by dci-control-server and python-dciclient',
    long_description=readme,
    install_requires=[],
    url='https://github.com/redhat-cip/python-dciauth',
    license='Apache v2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Security :: Cryptography'
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
import hashlib
import hmac


class Signature(object):
    def __init__(self, request, now=None):
        """
        dciauth Signature object implementing AWS HMAC version 4 mechanism.

        :param request (AuthRequest): dciauth AuthRequest object
        :param now (datetime): datetime used in Signature algorithm (default: datetime.utcnow())
        """

        self.request = request
        self.now = now or datetime.datetime.utcnow()
        self.dci_date = self.now.strftime('%Y%m%d')
        self.dci_datetime_format = '%Y%m%dT%H%M%SZ'
        self.dci_datetime = self.now.strftime(self.dci_datetime_format)
        self.dci_datetime_header = 'dci-datetime'

    def generate_headers(self, client_type, client_id, secret):
        """
        generate_headers is used to generate the headers automatically for your http request

        :param client_type (str): remoteci or feeder
        :param client_id (str): remoteci or feeder id
        :param secret (str): api secret
        :return: Authorization headers (dict)
        """

        self.request.add_header(self.dci_datetime_header, self.dci_datetime)
        signature = self._sign(secret)
        return self.request.build_headers(client_type, client_id, signature)

    def is_valid(self, secret):
        signature = self._sign(secret)
        client_signature = self.request.get_client_info()['signature']
        return self._equals(signature, client_signature)

    def is_expired(self):
        timestamp = self.request.headers.get(self.dci_datetime_header)
        if timestamp:
            timestamp = datetime.datetime.strptime(timestamp, self.dci_datetime_format)
            return abs(self.now - timestamp) > datetime.timedelta(days=1)
        return False

    def _create_canonical_request(self):
        payload_hash = hashlib.sha256(self.request.get_payload_string().encode('utf-8')).hexdigest()
        return """{method}
{endpoint}
{query_string}
{headers_string}
{signed_headers}
{payload_hash}""".format(method=self.request.method,
                         endpoint=self.request.endpoint,
                         query_string=self.request.get_query_string(),
                         headers_string=self.request.get_headers_string(),
                         signed_headers=self.request.get_signed_headers_string(),
                         payload_hash=payload_hash)

    def _create_string_to_sign(self):
        canonical_request = self._create_canonical_request()
        canonical_request_hash = hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
        return """{dci_algorithm}
{dci_datetime}
{dci_date}
{canonical_request_hash}""".format(dci_algorithm=self.request.algorithm,
                                   dci_datetime=self.dci_datetime,
                                   dci_date=self.dci_date,
                                   canonical_request_hash=canonical_request_hash)

    def _sign(self, secret):
        string_to_sign = self._create_string_to_sign()
        signing_key = hmac.new(
            secret.encode('utf-8'),
            self.dci_date.encode('utf-8'),
            hashlib.sha256
        ).digest()
        return hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

    @staticmethod
    def _equals(client_signature, header_signature):
        return hmac.compare_digest(
            client_signature.encode('utf-8'),
            header_signature.encode('utf-8')
        )
